
; Copyright (c) 2005-2020 Intel Corporation
;
; Licensed under the Apache License, Version 2.0 (the "License");
; you may not use this file except in compliance with the License.
; You may obtain a copy of the License at
;
;     http:
;
; Unless required by applicable law or agreed to in writing, software
; distributed under the License is distributed on an "AS IS" BASIS,
; WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
; See the License for the specific language governing permissions and
; limitations under the License.

EXPORTS

; frontend.cpp
scalable_calloc
scalable_free
scalable_malloc
scalable_realloc
scalable_posix_memalign
scalable_aligned_malloc
scalable_aligned_realloc
scalable_aligned_free
scalable_msize
scalable_allocation_mode
scalable_allocation_command
__TBB_malloc_safer_free
__TBB_malloc_safer_realloc
__TBB_malloc_safer_msize
__TBB_malloc_safer_aligned_msize
__TBB_malloc_safer_aligned_realloc
; memory pool stuff
?pool_create@rml@@YAPEAVMemoryPool@1@_JPEBUMemPoolPolicy@1@@Z
?pool_create_v1@rml@@YA?AW4MemPoolError@1@_JPEBUMemPoolPolicy@1@PEAPEAVMemoryPool@1@@Z
?pool_destroy@rml@@YA_NPEAVMemoryPool@1@@Z
?pool_malloc@rml@@YAPEAXPEAVMemoryPool@1@_K@Z
?pool_free@rml@@YA_NPEAVMemoryPool@1@PEAX@Z
?pool_reset@rml@@YA_NPEAVMemoryPool@1@@Z
?pool_realloc@rml@@YAPEAXPEAVMemoryPool@1@PEAX_K@Z
?pool_aligned_realloc@rml@@YAPEAXPEAVMemoryPool@1@PEAX_K2@Z
?pool_aligned_malloc@rml@@YAPEAXPEAVMemoryPool@1@_K1@Z
?pool_identify@rml@@YAPEAVMemoryPool@1@PEAX@Z
?pool_msize@rml@@YA_KPEAVMemoryPool@1@PEAX@Z

